"""
Sky subdivision after:
Bourgeois, Denis & Reinhart, C. & Ward, G. (2008). Standard daylight coefficient model for dynamic daylighting simulations. Building Research and Information - BUILDING RES INFORM. 36. 10.1080/09613210701446325.
Paper openly available on Researchgate: https://www.google.com/url?sa=t&source=web&rct=j&opi=89978449&url=https://www.researchgate.net/publication/228683789_Standard_daylight_coefficient_model_for_dynamic_daylighting_simulations&ved=2ahUKEwjp96OWi_iLAxWj8LsIHea6LUMQFnoECAoQAQ&usg=AOvVaw060IkWjmPP1qEXHvIA08C6

"""

import cProfile
import logging
import math
import numpy as np
import os
import pandas as pd
from pathlib import Path
from matplotlib import pyplot as plt
import matplotlib
matplotlib.use('TkAgg')

from MODULES.conversion_functions import sph_to_cart

logger = logging.getLogger(__name__)


def fibonacci_half_sphere(samples=18):
    """
    Function computing a number of direction sampling a virtual upper half sphere with a center at 0,0,0

    Parameters:
        samples (int): Number of directions

    Returns:
       Directions (np.array): Matrix (samples x 3) providing the direction
    """
    phi = np.pi * (3. - np.sqrt(5.))
    i = np.linspace(0, samples - 1, num=samples)
    yp = (1 - i / float(samples - 1))
    radius = np.sqrt(1 - yp ** 2)
    theta = phi * i
    xp = np.cos(theta) * radius
    zp = np.sin(theta) * radius
    return np.column_stack([xp, zp, yp])


class ReinhartSky:

    def __init__(self, MF=1):
        self.MF = MF

        self.reinhart_sky, self.reinhart_num_total = self.get_reinhart(MF=MF)

        self.build_patches()

        # Add the top patch
        self.add_top_patch()

        # Compute solid angles of all patches
        self.check_sum_solid_angles()

        # Compute surface areas of all patches
        self.compute_surface_areas()

        # Compute x, y, z coordinates of unit vectors pointing towards the sky patches
        self.get_patches_xyz()

        # Compute cos(zenith angle), used in irradiance computations
        self.compute_cos_zenith()

    def get_original_tregenza(self):
        original_tregenza_segments = pd.DataFrame()

        original_tregenza_segments['tregenza_row'] = np.arange(7)
        original_tregenza_segments['num_segments'] = np.array(
            [30, 30, 24, 24, 18, 12, 6])

        differential_altitude = 12  # [deg]
        original_tregenza_segments['altitude_deg'] = (np.arange(
            7) + 0.5) * differential_altitude

        original_tregenza_segments['diff_azimuth'] = np.array(
            [12, 12, 15, 15, 20, 30, 60])

        return original_tregenza_segments

    def get_corresp_tregenza145_row(self, F, MF):
        self.alpha = 90 / ((2 ** F * 7) + 1 / 2)  # [deg] Differential altitude

        self.num_rows = 2**F * 7

        reinhart_rows = np.arange(self.num_rows, dtype=int)
        tregenza_rows = np.floor(reinhart_rows/MF)

        return reinhart_rows, tregenza_rows

    def get_reinhart(self, MF, _verbose=False):
        F = np.log2(MF)

        original_tregenza_segments = self.get_original_tregenza()

        reinhart_sky = pd.DataFrame()
        reinhart_rows, tregenza_rows = self.get_corresp_tregenza145_row(
            F, MF)
        reinhart_sky['row'] = reinhart_rows
        reinhart_sky['altitude_deg'] = (reinhart_rows + 0.5) * self.alpha
        reinhart_sky['tregenza_row'] = tregenza_rows.astype(int)
        reinhart_sky = pd.merge(reinhart_sky, original_tregenza_segments[
            ['tregenza_row', 'num_segments']], on='tregenza_row', how='left')
        reinhart_sky = reinhart_sky.rename(
            columns={'num_segments': 'num_tregenza_segments'})

        reinhart_sky['num_segments'] = reinhart_sky[
                                           'num_tregenza_segments'] * round(
            2 ** F)

        reinhart_num_total = round(4 ** F * 144) + 1
        Ivanova_num_total = 144 * MF ** 2 + 1
        num_segments_checksum = reinhart_sky['num_segments'].sum() + 1

        # assert reinhart_num_total == num_segments_checksum
        diff1 = reinhart_num_total - Ivanova_num_total
        diff = num_segments_checksum - reinhart_num_total

        if _verbose or diff or diff1:
            print('\n', 5 * '#', f'{MF=} ; {F=}', 5 * '#')
            if not (diff) and not (diff1):
                print('Passed')

        if diff1:
            print('Mismatch in theoretical number of patches !!')
            print(f'Calc Reinhart = {reinhart_num_total}')
            print(f'Calc Ivanova et Gueymard = {Ivanova_num_total}')

        if diff != 0:
            print('Mismatch in implementation !!!')
        if diff > 0:
            print(f'{diff} extra patch(es)')
        elif diff < 0:
            print(f'{diff} missing patch(es)')

        return reinhart_sky, reinhart_num_total

    def build_patches(self):
        self.patch_id = 0
        self.cols = ['row', 'patch_id', 'patch_id_in_row', 'd_el', 'd_az', 'el',
                     'az', 'solid_angle_sr']

        # temporary list that will be appended to (dictionaries) in the loop,
        # then converted to Dataframe after the loop
        temp_list = []
        for row in self.reinhart_sky['row']:
            d_az = 360 / \
                   self.reinhart_sky[self.reinhart_sky['row'] == row][
                       'num_segments'].values[
                       0]
            el = self.alpha * (row + 1 / 2)  # [deg]

            for segment in range(self.reinhart_sky['num_segments'].iloc[row]):
                az = d_az * (segment + 1 / 2)  # [deg]
                solid_angle = self.compute_solid_angle_from_elev_azim(el,
                                                                      d_az)  # [sr]

                # Dictionary that will be appended to temp_list
                dict = {'row': row, 'patch_id': self.patch_id, 'patch_id_in_row': segment,
                        'd_el': self.alpha, 'd_az': d_az, 'el': el,
                     'az': az, 'solid_angle_sr': solid_angle}

                temp_list.append(dict)

                self.patch_id += 1

            # Convert the list of dictionaries to DataFrame
            self.reinhart_patches = pd.DataFrame.from_records(temp_list)

    def add_top_patch(self):
        top_row = max(self.reinhart_sky['row']) + 1
        top_patch_solid_angle = self.compute_solid_angle_cone(self.alpha / 2)
        top_patch = pd.DataFrame(
            [[top_row, self.patch_id, 0, self.alpha / 2, 360, 90, 0, top_patch_solid_angle]],
            columns=self.cols)
        self.reinhart_patches = pd.concat([self.reinhart_patches, top_patch],
                                     ignore_index=True)


    def compute_solid_angle_from_elev_azim(self, el, d_az):
        """

        Args:
            el: elevation at the center of the patch [deg]
            d_el: differential elevation of the patch [deg]
            d_az: differential azimuth of the patch [deg]

        Returns:
            solid_angle: solid angle of the patch [sr]
        """
        # inputs in deg, shape 2x1

        # convert to rad
        el = np.deg2rad(el)
        d_el = np.deg2rad(self.alpha)
        d_az = np.deg2rad(d_az)

        # compute the solid angle [sr]
        solid_angle = np.cos(el) * d_el * d_az

        # print(dw)

        return solid_angle

    def compute_solid_angle_cone(self, half_apex_angle):
        """

        Args:
            half_apex_angle: in degrees

        Returns:
            solid angle of a cone
        """
        half_apex_angle = np.deg2rad(half_apex_angle)
        solid_angle = 2 * np.pi * (1 - np.cos(half_apex_angle))

        return solid_angle

    def compute_surface_areas(self):
        # Compute surf area of each patch
        self.reinhart_patches['Normalized surf area'] = self.reinhart_patches['solid_angle_sr'] / (
                2 * np.pi)
        sum_surf_areas = self.reinhart_patches['Normalized surf area'].sum()

        logger.debug(f'Sum of surface areas = {sum_surf_areas:.4g}')

    def check_sum_solid_angles(self):
        # Check that the sum of solid angles is correct
        sum_solid_angles = self.reinhart_patches['solid_angle_sr'].sum()
        sum_solid_angles_div_pi = sum_solid_angles / np.pi
        logger.debug(f'Sum of solid angles = {sum_solid_angles_div_pi:.4g} * pi')
        if not math.isclose(sum_solid_angles, 2*np.pi, rel_tol=5e-3):
            logger.warning('Sum of solid angles seems off, check the sky discretization scheme')
            logger.warning(f'Sum of solid angles = {sum_solid_angles_div_pi:.6g} * pi')

    def get_patches_xyz(self):
        # Compute xyz coords on a unit sphere (actually half sphere since it's the sky dome)
        self.reinhart_patches['x'], \
        self.reinhart_patches['y'], \
        self.reinhart_patches['z'] = sph_to_cart(units='deg',
                                                 azimut=self.reinhart_patches['az'],
                                                 elev=self.reinhart_patches['el'])

    def compute_cos_zenith(self):
        self.reinhart_patches['cos(z)'] = np.cos(np.deg2rad(90-self.reinhart_patches['el']))

    def test_MFs(self):
        MFs = np.arange(1, 33)
        for MF in MFs:
            _, _ = self.get_reinhart(MF=MF, _verbose=False)

    def scatter_3D(self, text_id=False):
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')

        ax.scatter(self.reinhart_patches['x'],
                   self.reinhart_patches['y'],
                   self.reinhart_patches['z'], marker='.', c='k')

        if text_id:
            for idx in range(len(self.reinhart_patches)):
                patch = self.reinhart_patches.iloc[idx]
                ax.text(patch['x'], patch['y'], patch['z'],
                        s=f"{patch['patch_id']:.0f}")

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        title = f'{self.reinhart_num_total} patches (MF:{self.MF})'
        ax.set_title(title)
        ax.set_box_aspect([1.0, 1.0, 1/2])
        plt.show()

    def scatter_2D(self, text_id=False):
        fig = plt.figure()
        ax = fig.add_subplot()

        ax.scatter(self.reinhart_patches['x'], self.reinhart_patches['y'],
                   c=self.reinhart_patches['patch_id'])

        if text_id:
            for idx in range(len(self.reinhart_patches)):
                patch = self.reinhart_patches.iloc[idx]
                ax.text(patch['x'], patch['y'], f"{patch['patch_id']:.0f}")

        ax.set_xlabel('X')
        ax.set_ylabel('Y')

        ax.set_aspect('equal', adjustable='box')

        plt.show()


class CIEStandardSky:
    """
    Generate a standard radiance distribution among the 15 CIE General Skies.
    """

    def __init__(self, discrete_sky,
                 sun_az, sun_el,
                 sky_type=5):

        self.sky = discrete_sky

        self.az = discrete_sky['az'].values
        self.el = discrete_sky['el'].values

        self.az_s = sun_az  # [°] Sun azimuth (145° is south-east)
        self.el_s = sun_el  # [°]  Sun elevation

        self.sky_type = int(sky_type)

        self.rel_radiance_distribution = self.compute_rel_radiance()


    def compute_sph_angular_dist(self, Z, Z_s, az, az_s):
        """

        :param Z: Sky patch zenith angle [deg]
        :param Z: float
        :param Z_s: Solar zenith angle [deg]
        :param Z_s: float
        :param az: Sky patch azimuth angle from north [deg]
        :param az: float
        :param az_s: Solar azimuth angle from north [deg]
        :param az_s: float
        :return: chi = Spherical angular distance [rad]
        """

        # Convert all angles from degrees to radians
        Z = np.deg2rad(Z)
        Z_s = np.deg2rad(Z_s)
        az = np.deg2rad(az)
        az_s = np.deg2rad(az_s)

        # Absolute difference in azimuth angles
        Az = np.abs(az - az_s)

        # Angular distance chi
        chi = np.arccos(
            np.cos(Z_s) * np.cos(Z) + np.sin(Z_s) * np.sin(Z) * np.cos(Az))

        return chi

    def gradation_fun(self, a, b, Z):
        """
        Gradation function. It is not used as is but vectorized in compute_gradation().

        :param a: Horizon-zenith gradient (gradation parameter) [-]
        :type a: float
        :param b: Gradient intensity (gradation parameter) [-]
        :type b: float
        :param Z: Zenith angle [deg]
        :type Z: float
        :return: gradation function value [-]
        """
        if math.isclose(Z, np.pi / 2):
            return 1
        else:
            return 1 + a * np.exp(b / np.cos(Z))

    def compute_gradation(self, a, b, Z):
        """
        Vectorizes and computes the gradation function

        :param a: Horizon-zenith gradient (gradation parameter) [-]
        :type a: float
        :param b: Gradient intensity (gradation parameter) [-]
        :type b: float
        :param Z: Zenith angle [deg]
        :type Z: Nx1 array of float
        :return: gradation function [-], same shape as Z
        """

        v_gradation = np.vectorize(self.gradation_fun)

        Z = np.deg2rad(Z)

        return v_gradation(a, b, Z)

    def scattering_indicatrix_fun(self, c, d, e, x):
        """
        Compute the scattering indicatrix value.

        :param c: Circumsolar intensity (scattering parameter) [-]
        :type c: float
        :param d: Circumsolar radius (scattering parameter) [-]
        :type d: float
        :param e: Backscattering effect (scattering parameter) [-]
        :type e: float
        :param x: angle (spherical angular distance or zenith, i.e. 0) [rad]
        :type x: array of float or float
        :return: Scattering indicatrix value [-], same shape as x
        """
        return 1 + c * (np.exp(d * x) - np.exp(d * np.pi / 2)) + e * (
            np.cos(x)) ** 2

    def get_sky_params(self, standard_sky):
        """

        :param standard_sky: Dataframe of length 1 with standard sky information
        :return a: Horizon-zenith gradient (gradation parameter) [-]
        :return b: Gradient intensity (gradation parameter) [-]
        :return c: Circumsolar intensity (scattering parameter) [-]
        :return d: Circumsolar radius (scattering parameter) [-]
        :return e: Backscattering effect (scattering parameter) [-]
        """

        a = standard_sky['a'].values[0]
        b = standard_sky['b'].values[0]
        c = standard_sky['c'].values[0]
        d = standard_sky['d'].values[0]
        e = standard_sky['e'].values[0]

        return a, b, c, d, e

    def relative_radiance_fun(self, sky_type=None, az=None, el=None):
        """

        :return: relative quantity (radiance or luminance) with respect to
                 value at the zenith
        """

        # Read the csv containing the descriptions of the 15 standard skies

        pase_dir_path = Path(__file__).parents[2]
        standard_skies = pd.read_csv(os.path.join(pase_dir_path,
                                                  'INPUTS',
                                                  'CIE_standard_skies.csv'))

        # Extract the parameters of the desired sky type
        if sky_type is None:
            standard_sky = standard_skies.loc[standard_skies['Type'] == self.sky_type]
        else:
            standard_sky = standard_skies.loc[standard_skies['Type'] == sky_type]

        # Sky model parameters
        a, b, c, d, e = self.get_sky_params(standard_sky)

        # Elevation to zenith (Sun)
        Z_s = 90 - self.el_s  # [deg]

        # Elevation to zenith (sky patch)
        if el is None:
            Z = 90 - self.el  # [deg]
        else:
            Z = 90 - el  # [deg]


        # Spherical angular dist
        if az is None:
            chi = self.compute_sph_angular_dist(Z, Z_s, self.az, self.az_s)
        else:
            chi = self.compute_sph_angular_dist(Z, Z_s, az, self.az_s)

        # Gradation
        phi = self.compute_gradation(a, b, Z)
        phi_zenith = self.compute_gradation(a, b, Z=0)  # TODO do not recompute each time !

        # Scattering indicatrix
        indicatrix = self.scattering_indicatrix_fun(c, d, e, chi)
        indicatrix_zenith = self.scattering_indicatrix_fun(c, d, e, np.deg2rad(Z_s))  # TODO do not recompute each time !

        # Relative luminance/radiance (called "rel_quantity" for genericity)
        rel_quantity = phi * indicatrix / (phi_zenith * indicatrix_zenith)

        return rel_quantity

    def compute_rel_radiance(self, sky_type=None, az=None, el=None):
        v_radiance_fun = np.vectorize(self.relative_radiance_fun)
        return v_radiance_fun(sky_type, az, el)


if __name__ == '__main__':

    profile_flag = False

    if profile_flag:
        MFs = [1, 2]
        for MF in MFs:
            profile_output = os.path.join('..', '..', 'OUTPUTS', 'profiling', f'cprofile_output_sky_model_MF{MF}.prof')
            cProfile.run(f"ReinhartSky(MF={MF})", profile_output, sort='tottime')
    else:
        MF = 1

        sky = ReinhartSky(MF=MF)
        if MF == 1:
            text_id = True
        else:
            text_id = False

        if MF < 8:
            sky.scatter_2D(text_id=text_id)
            sky.scatter_3D(text_id=text_id)

    sky.reinhart_patches['Relative radiance distribution'] = CIEStandardSky(sky.reinhart_patches,
                               sun_az=147.67, sun_el=38.02, sky_type=15).rel_radiance_distribution

    DHI = 100  # [W/m²] arbitrary value for example's sake
    integral_rel_sky_radiance = (sky.reinhart_patches['Relative radiance distribution']
                                 * sky.reinhart_patches['solid_angle_sr']).sum()
    zenith_radiance = DHI/integral_rel_sky_radiance
    sky.reinhart_patches['Absolute radiance distribution'] = (sky.reinhart_patches['Relative radiance distribution']
                                                              * zenith_radiance)

